/***********************************************************************

 SV1Codec command line interface

 This file is a part of SV1Codec.
 SV1Codec is Copyright (C) 2002 by Marcin Grzegorczyk.

 Permission is granted to copy and redistribute this file under the
 terms of the Open Source MIT License. For more information please refer
 to the file sv1codec.c or to the SV1Codec documentation.

***********************************************************************/

#include "sv1ccli.h"
#include "sv1cmisc.h"

#include <string.h>
#include <ctype.h>
#include <assert.h>




static int CL_GetOptionArgument(
  const char **const pp,            /* parsing point (just after the option) */
  const char *const *const argv,    /* current command line argument */
  const char **const opargp)        /* will point to the option's argument */
/*
 For an option that accepts an argument, get that argument
 and return it in '*opargp' while advancing the parsing point
 appropriately.

 Return the number of following command line arguments to skip,
 or a negative value if no argument to the option can be found.
*/
{
  if (**pp == '=' || **pp == ':') {
    /* parameter follows the switch */
    *opargp = *pp + 1;
    *pp = strchr(*pp, '\0');  /* skip to the end of argument */
    return 0;
  }
  else if (argv[1]) {
    /* name in the next argument (after whitespace on the command line) */
    *opargp = argv[1];
    return 1;
  }
  else {
    /* Error: A filename must follow this argument */
    return -1;
  }
}



struct sCL_CmdOptionInfo {
  const char *s;            /* Pointer to the option (or call it a switch...) */
  const char *arg;          /* Pointer to the optional argument */
  size_t sl;                /* Length of the option */
};


typedef int tCL_CmdOptionHandler(struct sCL_ProgParams *const pparp, const struct sCL_CmdOptionInfo *const optinfop);


static tCL_CmdOptionHandler CL_OH_IOFile;
static tCL_CmdOptionHandler CL_OH_Decode;
static tCL_CmdOptionHandler CL_OH_Encode;
static tCL_CmdOptionHandler CL_OH_Chksum;
static tCL_CmdOptionHandler CL_OH_Force;
static tCL_CmdOptionHandler CL_OH_MaximumCompression;
static tCL_CmdOptionHandler CL_OH_SaveExtension;
static tCL_CmdOptionHandler CL_OH_SaveLetter;
static tCL_CmdOptionHandler CL_OH_TitleDat;
static tCL_CmdOptionHandler CL_OH_TTVersion;
static tCL_CmdOptionHandler CL_OH_FileNameBits;
static tCL_CmdOptionHandler CL_OH_URLStyleNames;
static tCL_CmdOptionHandler CL_OH_Help;
static tCL_CmdOptionHandler CL_OH_GameTitle;
static tCL_CmdOptionHandler CL_OH_NameEscapeStyle;
static tCL_CmdOptionHandler CL_OH_MessageOutputMode;


static const struct sCL_CmdOptions {
  const char *option;
  Bool takes_arg;
  tCL_CmdOptionHandler *handler;
} CL_option_table[] = {
  {"i",         1, CL_OH_IOFile},
  {"o",         1, CL_OH_IOFile},
  {"d",         0, CL_OH_Decode},
  {"e",         0, CL_OH_Encode},
  {"c",         0, CL_OH_Chksum},
  {"f",         0, CL_OH_Force},
  {"m",         0, CL_OH_MaximumCompression},
  {"sv0",       0, CL_OH_SaveExtension},
  {"sv1",       0, CL_OH_SaveExtension},
  {"sv2",       0, CL_OH_SaveExtension},
  {"ss0",       0, CL_OH_SaveExtension},
  {"ss1",       0, CL_OH_SaveExtension},
  {"trt",       0, CL_OH_SaveLetter},
  {"trc",       0, CL_OH_SaveLetter},
  {"trh",       0, CL_OH_SaveLetter},
  {"try",       0, CL_OH_SaveLetter},
  {"trp",       0, CL_OH_SaveLetter},
  {"td",        0, CL_OH_TitleDat},
  {"ttd",       0, CL_OH_TTVersion},
  {"tto",       0, CL_OH_TTVersion},
  {"ttpd",      0, CL_OH_TTVersion},
  {"7",         0, CL_OH_FileNameBits},
  {"8",         0, CL_OH_FileNameBits},
  {"u",         0, CL_OH_URLStyleNames},
  {"h",         0, CL_OH_Help},
  {"?",         0, CL_OH_Help},
  {"name",      1, CL_OH_GameTitle},
  {"title",     1, CL_OH_GameTitle},
  {"xca",       0, CL_OH_NameEscapeStyle},
  {"xcu",       0, CL_OH_NameEscapeStyle},
  {"xch",       0, CL_OH_NameEscapeStyle},
  {"xcp",       0, CL_OH_NameEscapeStyle},
  {"stdmsg",    0, CL_OH_MessageOutputMode},
  {"stdout",    0, CL_OH_MessageOutputMode},
  {"stderr",    0, CL_OH_MessageOutputMode},
  {NULL,        0, NULL}
};




static int CL_OH_IOFile(struct sCL_ProgParams *const pparp, const struct sCL_CmdOptionInfo *const optinfop)
{
  assert(optinfop->sl == 1);
  assert(strchr("io", tolower((unsigned char)*optinfop->s)));
  assert(optinfop->arg);
  *(tolower((unsigned char)*optinfop->s) == 'o' ? &pparp->fi.outpname : &pparp->fi.infname) = optinfop->arg;
  return 0;
}



#ifdef HAVE_PRAGMA_ARGSUSED
#pragma argsused
#endif
static int CL_OH_Decode(struct sCL_ProgParams *const pparp, PARAMETER_UNUSED(const struct sCL_CmdOptionInfo *const optinfop))
{
  pparp->op.operation = OP_DECODE;
  return 0;
}



#ifdef HAVE_PRAGMA_ARGSUSED
#pragma argsused
#endif
static int CL_OH_Encode(struct sCL_ProgParams *const pparp, PARAMETER_UNUSED(const struct sCL_CmdOptionInfo *const optinfop))
{
  pparp->op.operation = OP_ENCODE;
  return 0;
}



#ifdef HAVE_PRAGMA_ARGSUSED
#pragma argsused
#endif
static int CL_OH_Chksum(struct sCL_ProgParams *const pparp, PARAMETER_UNUSED(const struct sCL_CmdOptionInfo *const optinfop))
{
  pparp->op.operation = OP_CHKSUM;
  return 0;
}



#ifdef HAVE_PRAGMA_ARGSUSED
#pragma argsused
#endif
static int CL_OH_Force(struct sCL_ProgParams *const pparp, PARAMETER_UNUSED(const struct sCL_CmdOptionInfo *const optinfop))
{
  pparp->op.ignore_errors = 1;
  return 0;
}



#ifdef HAVE_PRAGMA_ARGSUSED
#pragma argsused
#endif
static int CL_OH_MaximumCompression(struct sCL_ProgParams *const pparp, PARAMETER_UNUSED(const struct sCL_CmdOptionInfo *const optinfop))
{
  pparp->op.rle_code_min = -128;
  pparp->op.rle_code_max = 128;
  return 0;
}



static int CL_OH_SaveExtension(struct sCL_ProgParams *const pparp, const struct sCL_CmdOptionInfo *const optinfop)
{
  int i;
  assert(optinfop->sl == 3);
  for (i = 0; i < 3; i++) pparp->fi.save_ext[i] = (char)toupper((unsigned char)optinfop->s[i]);
  pparp->op.title_dat = 0;  /* either fish or aquarium :-P */
  pparp->op.title_dat_forced = 1;
  return 0;
}



static int CL_OH_SaveLetter(struct sCL_ProgParams *const pparp, const struct sCL_CmdOptionInfo *const optinfop)
{
  assert(optinfop->sl == 3);
  pparp->fi.save_letter = (char)toupper((unsigned char)optinfop->s[2]);
  return 0;
}



#ifdef HAVE_PRAGMA_ARGSUSED
#pragma argsused
#endif
static int CL_OH_TitleDat(struct sCL_ProgParams *const pparp, PARAMETER_UNUSED(const struct sCL_CmdOptionInfo *const optinfop))
{
  pparp->op.title_dat = 1;
  pparp->op.title_dat_forced = 1;
  return 0;
}



static int CL_OH_TTVersion(struct sCL_ProgParams *const pparp, const struct sCL_CmdOptionInfo *const optinfop)
{
  assert(optinfop->sl >= 3);
  switch (tolower((unsigned char)optinfop->s[2])) {
    default:
      #define TTVersionCLOptionBad 0
      assert(TTVersionCLOptionBad);
    case 'p':
      assert(optinfop->sl == 4);
      assert(tolower((unsigned char)optinfop->s[3]) == 'd');
      pparp->op.sv_version = SV_VER_TTp_DEMO;
      break;
    case 'o':
      assert(optinfop->sl == 3);
      pparp->op.sv_version = SV_VER_TTO;
      break;
    case 'd':
      assert(optinfop->sl == 3);
      pparp->op.sv_version = SV_VER_TTD;
  }
  pparp->op.sv_ver_forced = 1;
  return 0;
}



static int CL_OH_FileNameBits(struct sCL_ProgParams *const pparp, const struct sCL_CmdOptionInfo *const optinfop)
{
  assert(optinfop->sl == 1);
  assert(*optinfop->s == '7' || *optinfop->s == '8');
  pparp->op.use_8bit_filenames = (*optinfop->s == '8');
  return 0;
}



#ifdef HAVE_PRAGMA_ARGSUSED
#pragma argsused
#endif
static int CL_OH_URLStyleNames(struct sCL_ProgParams *const pparp, PARAMETER_UNUSED(const struct sCL_CmdOptionInfo *const optinfop))
{
  pparp->op.read_escape_modes |= FNAME_ESCAPE_URL;
  pparp->op.write_escape_mode = FNAME_ESCAPE_URL;
  return 0;
}



#ifdef HAVE_PRAGMA_ARGSUSED
#pragma argsused
#endif
static int CL_OH_Help(struct sCL_ProgParams *const pparp, PARAMETER_UNUSED(const struct sCL_CmdOptionInfo *const optinfop))
{
  pparp->help_requested = 1;
  return 0;
}



static int CL_OH_GameTitle(struct sCL_ProgParams *const pparp, const struct sCL_CmdOptionInfo *const optinfop)
{
  assert(optinfop->arg);
  pparp->game_title = optinfop->arg;
  return 0;
}



static int CL_OH_NameEscapeStyle(struct sCL_ProgParams *const pparp, const struct sCL_CmdOptionInfo *const optinfop)
{
  unsigned mode;

  assert(optinfop->sl == 3);
  switch (tolower((unsigned char)optinfop->s[2])) {
    default:
      #define FileNameStyleCLOptionBad 0
      assert(FileNameStyleCLOptionBad);
    case 'a': mode = FNAME_ESCAPE_AT; break;
    case 'u': mode = FNAME_ESCAPE_URL; break;
    case 'h': mode = FNAME_ESCAPE_HASH; break;
    case 'p': mode = FNAME_ESCAPE_POSIX; break;
  }
  pparp->op.read_escape_modes |= mode;
  pparp->op.write_escape_mode = mode;
  return 0;
}



static int CL_OH_MessageOutputMode(struct sCL_ProgParams *const pparp, const struct sCL_CmdOptionInfo *const optinfop)
{
  assert(optinfop->sl > 3);
  if (StrnEQ(optinfop->s + 3, "out", 3)) pparp->msg_mode = MSG_MODE_STDOUT;
  else if (StrnEQ(optinfop->s + 3, "err", 3)) pparp->msg_mode = MSG_MODE_STDERR;
  else {
    assert(StrnEQ(optinfop->s + 3, "msg", 3));
    pparp->msg_mode = MSG_MODE_STDMSG;
  }
  return 0;
}



PUBLIC_DEFINITION
int CL_ParseCmdLine(struct sCL_ProgParams *const pparp, const char *const *argv)
/*
 Parse the command line given as the argv array (the argv[0] is assumed
 to be already skipped) and set various fields in the structure pointed to
 by 'pparp'.

 Return 0 on success, nonzero on failure.
*/
{
  int no_args = 0;  /* number of arguments that follow an option (a switch) */
  Bool no_more_options = 0;

  while (*argv) {
    const char *pp = *argv;  /* parsing point */

    if (!no_more_options && *pp == '-') {
      pp++;
      if (*pp == '-' && pp[1] == '\0') no_more_options = 1;
      else while (*pp) {
        /* Search for the current switch in the table */
        /* (now 'pp' points to the current switch) */
        const struct sCL_CmdOptions *sw = CL_option_table;

        for (;; sw++) {
          size_t optl;  /* length of the current switch in the table */

          if (sw->option == NULL) {
            /* None of the switches in the table matched */
            UI_UnrecognizedOption(pp, strlen(pp));
            return 1;
          }
          optl = strlen(sw->option);

          if (StrnEQ(sw->option, pp, optl)) {
            /* Switch found in the table */
            struct sCL_CmdOptionInfo optinfo;

            DEBUG_NULLINIT(optinfo.arg);
            optinfo.s = pp;
            optinfo.sl = optl;
            pp += optinfo.sl;  /* advance past the switch */
            if (sw->takes_arg) {
              int na = CL_GetOptionArgument(&pp, argv, &optinfo.arg);
              if (na < 0) {
                UI_MissingOptionArgument(optinfo.s, optinfo.sl);
                return 2;
              }
              /* update the number of arguments to skip later */
              if (na > no_args) no_args = na;
              assert(na <= 1);  /* currently no option takes more than 1 argument */
              assert(optinfo.arg != NULL);
            }
            else optinfo.arg = NULL;

            /* handle the switch */
            assert(sw->handler != NULL);
            if (sw->handler(pparp, &optinfo)) return 3;

            break;  /* switch found and handled, stop searching */
          }
        }
        /* search again unless the string has been exhausted */
      }
    }
    else {
      /* Not an option string -- must be a filename */
      if (pparp->fi.infname) {
        if (pparp->fi.outpname) {
          UI_TooManyFilenames();
          return 4;
        }
        pparp->fi.outpname = pp;
      }
      else pparp->fi.infname = pp;
    }

    /* advance to the next command line argument */
    argv += no_args + 1;
    no_args = 0;
  }
  return 0;
}

