/***********************************************************************

 SV1Codec RLE compression/decompression
 and various TT savegame format handling routines

 This file is a part of SV1Codec.
 SV1Codec is Copyright (C) 2002 by Marcin Grzegorczyk.
 See sv1codec.c for details.

 The code in this file is explicitly placed in the public domain.
 Use, abuse and redistribute freely.

***********************************************************************/

#include "sv1crle.h"
#include "sv1cmisc.h"

#include <assert.h>



PUBLIC_DEFINITION
uint16_t CalcTitleChecksum(
  const unsigned char *titlebuf,  /* pointer to the name of the savegame */
  size_t buflen)                  /* length of the title part; 39 or 47 for TTO or TTD savegames, respectively */
/*
 Calculate checksum of the title part.
 This function has been adapted from the Chris Becke's TTD Internals page
 (no longer existing).
*/
{
  uint16_t sum = 0;
  size_t i;

  for(i = 0; i < buflen; i++) {
    sum += titlebuf[i];
    sum = (uint16_t)((sum << 1) | (sum >> 15));  /* rotate 'sum' 1 bit to the left */
  }

  return (uint16_t)(sum ^ 0xAAAA);
}



PUBLIC_DEFINITION
void ChksumStep(const unsigned char c, uint32_t *const chksump)
/*
 File checksum calculation: process one byte
*/
{
  /* add 'c' to the lowest byte of the checksum */
  *chksump = (unsigned char)(*chksump + c) | (*chksump & 0xFFFFFF00uL);
  /* rotate checksum 3 bits to the left */
  *chksump = (*chksump << 3) | (*chksump >> 29);
}



PUBLIC_DEFINITION
uint32_t CalcFileChecksum(
  const unsigned char *data,    /* pointer to the buffer that contains the entire file */
  size_t length,                /* size of the buffer (checksum field excluded) */
  uint32_t adjustment)          /* final checksum adjustment */
/*
 Calculate file checksum.
 This function has been adapted from the Chris Becke's TTD Internals page.
*/
{
  uint32_t checksum = 0;
  size_t i;

  /* sequentially add all bytes to the checksum */
  for(i = 0; i < length; i++) ChksumStep(data[i], &checksum);

  /* make final adjustment */
  return checksum + adjustment;
}



PUBLIC_DEFINITION
int RLE_Decode(
  const unsigned char *pSrc,                  /* pointer to the source RLE data */
  const size_t size,                          /* the size of the source data */
  tRLEWriteResultCallback *const write_call)  /* callback to write the result, see the header */
/*
 Expand RLE-compressed data. This function has been adapted from
 the Chris Becke's page; the major difference is that it writes
 its output to a file stream, not to another buffer.

 Return 0 on success, -1 on an EOF condition (logical, eh?),
 or a positive integer on write error.
*/
{
  const unsigned char *const pEnd = pSrc + size;  /* pointer past the end of source data */
  int code;         /* for the RLE byte count code */
  unsigned char v;  /* repeated value (temporary) */

  /* Each iteration of this loop decodes one RLE sequence */
  while(pSrc < pEnd) {
    /* read byte count into 'code' */
    code = (signed char)*pSrc++;
    if (pSrc == pEnd) return -1;  /* RLE error - orphaned byte count */

    if(code < 0) {
      /* repeat the next byte (-code+1) times */
      v = *pSrc++;
      do if (write_call(&v, 1)) return 1; while(code++);
    }
    else {
      /* copy the next (code+1) bytes */
      code++;
      if (code > (pEnd - pSrc)) {
        /* RLE error - data stream truncated */
        /* copy the remaining bytes and report error */
        if (write_call(pSrc, (size_t)(pEnd - pSrc))) return 1;
        return -1;
      }
      /* if no error, write these bytes and adjust data pointer */
      if (write_call(pSrc, code)) return 1;
      pSrc += code;
    }
  }
  return 0;
}



PUBLIC_DEFINITION
int RLE_Encode(
  const unsigned char *pSrc,                  /* pointer to the source raw data */
  const size_t size,                          /* the size of the source data */
  tRLEWriteResultCallback *const write_call,  /* callback to write the result, see the header */
  uint32_t *const chksump,                    /* pointer to the checksum */
  const int code_min,                         /* minimum permitted RLE code */
  const int code_max)                         /* maximum permitted RLE code */
/*
 Compress data using TT's Run Length Encoding (RLE) scheme, writing the
 result to a file stream, and calculate the file checksum in the process.

 The permitted range for 'code_min' is -128 through -1; the permitted range
 for 'code_max' is 1 through 128. The default values for 'code_min' and
 'code_max' that replicate TT's behaviour are -124 and 124, respectively.
 Note that 'code_max' can be legally 128 even though the actual RLE code
 fits in one (signed) byte and thus cannot exceed 127.

 Return 0 on success, or a positive integer on write error.

 NOTE: this function assumes that the game title and title checksum fields
 have been already processed by the file checksum calculation routine.
*/
{
  const unsigned char *const pEnd = pSrc + size;  /* pointer past the end of source data */
  unsigned char v, c[2];  /* temporaries */
  int code;               /* for the RLE byte count code */

  /* Each iteration of this loop encodes one RLE sequence */
  while(pSrc < pEnd) {
    v = *pSrc++;  /* read the first byte */
    c[1] = v;     /* save it for later */
    if(pSrc < pEnd && *pSrc == v) {  /* two or more adjacent same-value bytes */
      /* determine how many bytes have the same value and can be RLEncoded */
      code = -1;
      while (++pSrc < pEnd && *pSrc == v && code > code_min) code--;
      assert(code < 0 && code >= -128);
      /* write: negative count byte ('code'), value byte ('v') */
      /* 'v' is already saved in 'c[1]' */
      c[0] = (unsigned char)code;
      if (write_call(c, 2)) return 1;
      /* process those 2 bytes for the checksum */
      ChksumStep(c[0], chksump);
      ChksumStep(c[1], chksump);
    }
    else {
      /* scan for the next two adjacent bytes with the same value */
      for (code = 0; ((pSrc + code) < pEnd && code < code_max); code++) {
        /* 'v' is the current byte, 'pSrc[code]' is the next byte */
        if (pSrc[code] == v) {          /* bytes have the same value */
                                        /* Note: can't happen when code == 0 */
          code--;                       /* don't include those bytes (they're to be RLEncoded) */
          break;
        }
        /* otherwise advance to the next byte */
        v = pSrc[code];
      }
      assert(code >= 0 && code <= 128);
      if (code >= 128) code = 127;  /* prevent overflow if full range is used */
      c[0] = (unsigned char)code;
      /* write: nonnegative count ('code'), (code+1) following bytes */
      /* Note that 'pSrc' has been advanced after reading the first byte, which went into 'c[1]' */
      if (write_call(c, 2)) return 1;
      if (code && write_call(pSrc, code)) return 1;
      /* process all those bytes for the checksum and advance 'pSrc' appropriately */
      ChksumStep(c[0], chksump);
      ChksumStep(c[1], chksump);
      while (code--) ChksumStep(*pSrc++, chksump);
    }
  }
  return 0;
}



PUBLIC_DEFINITION
enum eSVVersion DetectTTByTitle(const unsigned char *buf, const size_t len)
/*
 Detect whether a savegame is from TTO or TTD by the name buffer.
 Return an appropriate eSVVersion enumeration constant
 (SV_VER_UNKNOWN if the type cannot be recognized). Note that this
 function cannot distinguish between TTO and TTO Demo games,
 and returns SV_VER_TTO for both.
*/
{
  if (len < (TTx_MAX_TITLE_LENGTH + 2)) return SV_VER_UNKNOWN;
  /*
   The '8' here is:
   - title checksum = 2 bytes
   - minimum-length RLE sequence = 2 bytes
   - file checksum = 4 bytes
  */

  /* Check if the title checksum matches */
  if (GetLEInt16(buf + TTD_TITLE_LENGTH) == CalcTitleChecksum(buf, TTD_TITLE_LENGTH)) return SV_VER_TTD;
  if (GetLEInt16(buf + TTO_TITLE_LENGTH) == CalcTitleChecksum(buf, TTO_TITLE_LENGTH)) return SV_VER_TTO;
  return SV_VER_UNKNOWN;
}



PUBLIC_DEFINITION
enum eSVVersion DetectTTBySize(long flen, int *const ttdp)
/*
 Detect what version of TT a 'big' file belongs to by examining its length.
 Return an appropriate eSVVersion enumeration constant. If the 'big' file
 is a TTD uncompressed savegame, detect the TTDPatch extended array
 multiplier as well and return it via 'ttdp'.
*/
{
  if (flen == TTO_BIG_SIZE) return SV_VER_TTO;
  if (flen == TTp_BIG_SIZE) return SV_VER_TTp_DEMO;
  if (flen >= TTD_BIG_SIZE && flen <= (TTD_BIG_SIZE + TTD_BIG_XSIZE*(TTD_MAX_MOREVEHICLES - 1)) && ((flen - TTD_BIG_SIZE) % TTD_BIG_XSIZE) == 0) {
    *ttdp = (int)((flen - (TTD_BIG_SIZE - TTD_BIG_XSIZE)) / TTD_BIG_XSIZE);
    return SV_VER_TTD;
  }

  return SV_VER_UNKNOWN;
}



PUBLIC_DEFINITION
size_t TitleLengthFromVersion(enum eSVVersion sv_v)
{
  switch (sv_v) {
    default:
      #define TTTypeBad 0
      assert(TTTypeBad);
    case SV_VER_UNKNOWN:
      #define TTTypeMustBeKnown 0
      assert(TTTypeMustBeKnown);
    case SV_VER_TTO:
    case SV_VER_TTp_DEMO: return TTO_TITLE_LENGTH;
    case SV_VER_TTD: return TTD_TITLE_LENGTH;
  }
}



PUBLIC_DEFINITION
uint32_t ChecksumAdjFromVersion(enum eSVVersion sv_v)
{
  switch (sv_v) {
    default:
      #define TTTypeBad 0
      assert(TTTypeBad);
    case SV_VER_UNKNOWN:
      #define TTTypeMustBeKnown 0
      assert(TTTypeMustBeKnown);
    case SV_VER_TTO:      return TTO_FILE_CHECKSUM;
    case SV_VER_TTp_DEMO: return TTp_FILE_CHECKSUM;
    case SV_VER_TTD:      return TTD_FILE_CHECKSUM;
  }
}



PUBLIC_DEFINITION
int GetLandscapeStyle(const struct sBufferInfo *const buf, const int ttdp)
/*
 Get the landscape style (a.k.a. 'climate') of the uncompressed savegame
 loaded into 'buf'.

 Return the 'climate' byte (0=temperate, 1=sub-arctic, 2=sub-tropical,
 3=toyland), or a negative value if the landscape style cannot be
 determined because the TTDPatch vehicle array extension value, 'ttdp',
 is not set for some reason.
*/
{
  if (ttdp) {
    size_t offset = (size_t)TTDOffset(TTD_CLIMATE_OFFSET, ttdp);
    assert(offset < buf->len);  /* sanity check */
    return buf->ptr[offset];
  }
  else return -1;
}



PUBLIC_DEFINITION
int GetNumberOfPlayers(
  const struct sBufferInfo *const buf,
  enum eSVVersion sv_v,
  const int ttdp)
/*
 Get the number of players (0, 1 or 2) of the uncompressed savegame
 loaded into 'buf', or a negative value if it cannot be determined.
*/
{
  size_t offset;

  switch (sv_v) {
    default: return -1;
    case SV_VER_TTO:
      offset = TTO_PLAYERS_OFFSET;
      break;
    case SV_VER_TTD:
      offset = (size_t)TTDOffset(TTD_PLAYERS_OFFSET, ttdp);
      break;
  }
  assert((offset + 1) < buf->len);  /* sanity check */
  if (buf->ptr[offset] > 7) return 0;
  else return (buf->ptr[offset+1] > 7) ? 1 : 2;
}

